import io

import six
try:
    import numpy
    NUMPY_AVAILABLE = True
except ImportError:
    numpy = None
    NUMPY_AVAILABLE = False


from .base import BaseItertool


def iter_(obj):
    """A custom replacement for iter(), dispatching a few custom picklable
    iterators for known types.
    """
    if six.PY2:
        file_types = file,  # noqa
    if six.PY3:
        file_types = io.IOBase,
        dict_items = {}.items().__class__
        dict_values = {}.values().__class__
        dict_keys = {}.keys().__class__
        dict_view = (dict_items, dict_values, dict_keys)

    if isinstance(obj, dict):
        return ordered_sequence_iterator(list(obj.keys()))
    if isinstance(obj, file_types):
        return file_iterator(obj)
    if six.PY2:
        if isinstance(obj, (list, tuple)):
            return ordered_sequence_iterator(obj)
        if isinstance(obj, xrange):  # noqa
            return range_iterator(obj)
        if NUMPY_AVAILABLE and isinstance(obj, numpy.ndarray):
            return ordered_sequence_iterator(obj)
    if six.PY3 and isinstance(obj, dict_view):
        return ordered_sequence_iterator(list(obj))
    return iter(obj)


class range_iterator(BaseItertool):
    """A picklable range iterator for Python 2."""
    def __init__(self, xrange_):
        self._start, self._stop, self._step = xrange_.__reduce__()[1]
        self._n = self._start

    def __next__(self):
        if (self._step > 0 and self._n < self._stop or
                self._step < 0 and self._n > self._stop):
            value = self._n
            self._n += self._step
            return value
        else:
            raise StopIteration


class file_iterator(BaseItertool):
    """A picklable file iterator."""
    def __init__(self, f):
        self._f = f

    def __next__(self):
        line = self._f.readline()
        if not line:
            raise StopIteration
        return line

    def __getstate__(self):
        name, pos, mode = self._f.name, self._f.tell(), self._f.mode
        return name, pos, mode

    def __setstate__(self, state):
        name, pos, mode = state
        self._f = open(name, mode=mode)
        self._f.seek(pos)


class ordered_sequence_iterator(BaseItertool):
    """A picklable replacement for list and tuple iterators."""
    def __init__(self, sequence):
        self._sequence = sequence
        self._position = 0

    def __next__(self):
        if self._position < len(self._sequence):
            value = self._sequence[self._position]
            self._position += 1
            return value
        else:
            raise StopIteration



from abc import ABCMeta, abstractmethod
import six
from .product import product
from .base import BaseItertool


@six.add_metaclass(ABCMeta)
class IndexBased(BaseItertool):
    def __init__(self, iterable, r=None):
        self._pool = tuple(iterable)
        self._r = r if r is not None else len(self._pool)
        self._iter = self._construct_iter()

    def _construct_iter(self):
        return product(range(len(self._pool)), repeat=self._r)

    @abstractmethod
    def _valid_indices(self, indices):
        pass

    def __next__(self):
        indices = next(self._iter)
        while not self._valid_indices(indices):
            indices = next(self._iter)
        return tuple(self._pool[i] for i in indices)


class permutations(IndexBased):
    """permutations(iterable[, r]) --> permutations object

    Return successive r-length permutations of elements in the iterable.

    permutations(range(3), 2) --> (0,1), (0,2), (1,0), (1,2), (2,0), (2,1)
    """
    def _valid_indices(self, indices):
        return len(set(indices)) == self._r


@six.add_metaclass(ABCMeta)
class AbstractCombinations(IndexBased):
    def __init__(self, iterable, r):
        super(AbstractCombinations, self).__init__(iterable, r)

    def _valid_indices(self, indices):
        return sorted(indices) == list(indices)


class combinations_with_replacement(AbstractCombinations):
    """combinations_with_replacement(iterable, r) -->
    combinations_with_replacement object

    Return successive r-length combinations of elements in the iterable
    allowing individual elements to have successive repeats.
    combinations_with_replacement('ABC', 2) --> AA AB AC BB BC CC
    """
    pass


class combinations(AbstractCombinations):
    """combinations(iterable, r) --> combinations object

    Return successive r-length combinations of elements in the iterable.

    combinations(range(4), 3) --> (0,1,2), (0,1,3), (0,2,3), (1,2,3)
    """
    def _construct_iter(self):
        return permutations(range(len(self._pool)), self._r)
